# coding: utf-8
"""
These functions help you use hydrology analysis.
"""

import logging as _logging
import arcgis
from arcgis.geoprocessing._support import _execute_gp_tool
from arcgis.features import FeatureSet
from arcgis.features.geo._accessor import _is_geoenabled
from pandas import DataFrame

_log = _logging.getLogger(__name__)


def _evaluate_spatial_input(input_points):
    """
    Helper function to determine if the input is either a FeatureSet or Spatially Enabled DataFrame, and
    output to FeatureSet for subsequent processing.
    :param input_points: FeatureSet or Spatially Enabled DataFrame
    :return: FeatureSet
    """
    if isinstance(input_points, FeatureSet):
        return input_points

    elif isinstance(input_points, DataFrame) and _is_geoenabled(input_points):
        return input_points.spatial.to_featureset()

    elif isinstance(input_points, DataFrame) and not _is_geoenabled(input_points):
        raise Exception('input_points is a DataFrame, but does not appear to be spatially enabled. Using the <df>.spatial.set_geometry(col, sr=None) may help. (https://esri.github.io/arcgis-python-api/apidoc/html/arcgis.features.toc.html#arcgis.features.GeoAccessor.set_geometry)')

    else:
        raise Exception('input_points must be either a FeatureSet or Spatially Enabled DataFrame instead of {}'.format(type(input_points)))


def trace_downstream(input_points, point_id_field=None, source_database='Finest', generalize=False,
                     gis=None, future=False):
    """

    .. image:: _static/images/trace_downstream/trace_downstream.png

    The ``trace_downstream`` method delineates the downstream path from a specified location.
    Esri-curated elevation data is used to create an output polyline delineating the flow path
    downstream from the specified input location. This method accesses a service using multiple
    source databases which are available for different geographic areas and at different
    spatial scales.

    ==================     ====================================================================
    **Argument**           **Description**
    ------------------     --------------------------------------------------------------------
    input_points           Required FeatureSet or Spatially Enabled DataFrame
                           Points delineating the starting location to calculate the downstream
                           location from. See :ref:`Feature Input<FeatureInput>`.
    ------------------     --------------------------------------------------------------------
    point_id_field         Optional string. Field used to identify the feature from the source data. This is
                           useful for relating the results back to the original source data.
    ------------------     --------------------------------------------------------------------
    source_database        Optional string. Keyword indicating the source data that will be used in the
                           analysis. This keyword is an approximation of the spatial resolution
                           of the digital elevation model used to build the foundation
                           hydrologic  database. Since many elevation sources are distributed
                           with units of  arc seconds, this keyword is an approximation in
                           meters for easier understanding.


                           - Finest: Finest resolution available at each location from all
                             possible data sources.
                           - 10m: The hydrologic source was built from 1/3 arc second -
                             approximately 10 meter resolution, elevation data.
                           - 30m: The hydrologic source was built from 1 arc second -
                             approximately 30 meter resolution, elevation data.
                           - 90m: The hydrologic source was built from 3 arc second -
                             approximately 90 meter resolution, elevation data.

                           The default value is 'Finest'.
    ------------------     --------------------------------------------------------------------
    generalize             Optional boolean. Determines if the output downstream trace lines will be smoothed
                           into simpler lines.

                           The default value is False.
    ------------------     --------------------------------------------------------------------
    gis                    Optional GIS Object instance. If not provided as input, a GIS object instance logged into an
                           active portal with elevation helper services defined must already
                           be created in the active Python session. A GIS object instance can
                           also be optionally explicitly passed in through this parameter.
    ------------------     --------------------------------------------------------------------
    future                 Optional boolean. If True, the result will be a `GPJob` and results will be returned asynchronously.
    ==================     ====================================================================

    :return: FeatureSet

    .. code-block:: python

        # USAGE EXAMPLE: To trace downstream path from from the outlet points.
        path = trace_downstream(input_points=fs,
                                source_database='Finest',
                                generalize=False)
    """
    kwargs = locals()

    param_db = {
        "input_points": (FeatureSet, "InputPoints"),
        "point_id_field": (str, 'PointIDField'),
        "source_database": (str, 'SourceDatabase'),
        "generalize": (str, 'Generalize'),
        "output_trace_line": (FeatureSet, "Output Trace Line"),
    }

    return_values = [
        {"name": "output_trace_line", "display_name": "Output Trace Line", "type": FeatureSet},
    ]

    # use helper function to evaluate the input points and convert them, if necessary, to a FeatureSet
    input_fs = _evaluate_spatial_input(input_points)

    if input_fs.geometry_type != 'esriGeometryPoint':
        raise Exception('input_points FeatureSet must be point esriGeometryPoint, not {}'.format(input_fs.geometry_type))

    input_fields = input_fs.fields
    if point_id_field and point_id_field not in [f['name'] for f in input_fields] and len(input_fields):
        input_fields_str = ','.join(input_fields)
        raise Exception('The provided point_id_field {} does not appear to be in the input_points FeatureSet fields - {}'.format(point_id_field, input_fields_str))

    if source_database not in ['Finest', '10m', '30m', '90m']:
        raise Exception('source_database must be either "Finest", "10m", "30m", or "90m". {} does not appear to be one of these.'.format(source_database))

    if gis is None and arcgis.env.active_gis is None:
        raise Exception('GIS must be defined either by directly passing in a GIS object created using credentials, or one must already be created in the active Python session.')
    elif gis is None:
        gis = arcgis.env.active_gis

    url = gis.properties.helperServices.hydrology.url

    return _execute_gp_tool(gis, "TraceDownstream", kwargs, param_db, return_values, True, url, future=future)


def watershed(input_points, point_id_field=None, snap_distance=10, snap_distance_units='Meters',
              source_database='Finest', generalize=False, gis=None, return_snapped_points=True,
              future=False):
    """
    .. image:: _static/images/create_watersheds/create_watersheds.png

    The ``watershed`` is used to identify catchment areas based on a particular location you
    provide and ArcGIS Online Elevation data.


    ========================     ====================================================================
    **Argument**                 **Description**
    ------------------------     --------------------------------------------------------------------
    input_points                 Required FeatureSet or Spatially Enabled DataFrame. Points delineating the starting location to calculate the downstream
                                 location from. See :ref:`Feature Input<FeatureInput>`.
    ------------------------     --------------------------------------------------------------------
    point_id_field               Optional String. Field used to identify the feature from the source data. This is
                                 useful for relating the results back to the original source data.
    ------------------------     --------------------------------------------------------------------
    snap_distance                Optional float. The maximum distance to move the location of an input point.

                                 Interactive input points and documented gage locations may not exactly align with the stream location in the DEM.
                                 This parameter allows the task to move the point to a nearby location with the largest contributing area.

                                 The snap distance should always be larger than the source data resolution. By default, the snapping distance
                                 is calculated as the resolution of the source data multiplied by 5.

                                 The default value is 10.
    ------------------------     --------------------------------------------------------------------
    snap_distance_units          Optional String. The linear units specified for the snap distance.

                                 Choice list: ['Meters', 'Kilometers', 'Feet', 'Yards', 'Miles'].

                                 The default value is 'Meters'.
    ------------------------     --------------------------------------------------------------------
    source_database              Optional String. Keyword indicating the source data that will be used in the analysis.
                                 This keyword is an approximation of the spatial resolution of the
                                 digital elevation model used to build the foundation hydrologic
                                 database. Since many elevation sources are distributed with units of
                                 arc seconds, this keyword is an approximation in meters for easier
                                 understanding.

                                 * ``Finest``: Finest resolution available at each location from all possible data sources.
                                 * ``10m``: The hydrologic source was built from 1/3 arc second - approximately 10 meter resolution, elevation data.
                                 * ``30m``: The hydrologic source was built from 1 arc second - approximately 30 meter resolution, elevation data.
                                 * ``90m``: The hydrologic source was built from 3 arc second - approximately 90 meter resolution, elevation data.

                                 The default value is 'Finest'.
    ------------------------     --------------------------------------------------------------------
    generalize                   Optional boolean. Determines if the output downstream trace lines will be smoothed
                                 into simpler lines.

                                 The default value is False.
    ------------------------     --------------------------------------------------------------------
    gis                          Optional GIS Object instance. If not provided as input, a GIS object instance logged into an
                                 active portal with elevation helper services defined must already
                                 be created in the active Python session. A GIS object instance can
                                 also be optionally explicitly passed in through this parameter.
    ------------------------     --------------------------------------------------------------------
    return_snapped_points        Optional boolean. Determines if a point feature at the watershed’s pour point will be returned.
                                 If snapping is enabled, this might not be the same as the input point.

                                 The default value is True.
    ------------------------     --------------------------------------------------------------------
    future                       Optional boolean. If True, the result will be a `GPJob` and results will be returned asynchronously.
    ========================     ====================================================================

    :return:
        Result object comprised of two FeatureSets - one for watershed_area, and another for snapped_points

    .. code-block:: python

            # USAGE EXAMPLE: To identify catchment areas around Chennai lakes.
            lakes_watershed = watershed(input_points=lakes_fs,
                                        snap_distance=10,
                                        snap_distance_units='Meters',
                                        source_database='Finest',
                                        generalize=False)
    """

    kwargs = locals()

    param_db = {
        "input_points": (FeatureSet, "InputPoints"),
        "point_id_field": (str, 'PointIDField'),
        "snap_distance": (int, 'SnapDistance'),
        "snap_distance_units": (int, 'SnapDistanceUnits'),
        "source_database": (str, 'SourceDatabase'),
        "generalize": (str, 'Generalize'),
        "return_snapped_points": (str, 'ReturnSnappedPoints'),
        "watershed_area": (FeatureSet, "WatershedArea"),
        "snapped_points": (FeatureSet, "SnappedPoints")
    }

    return_values = [
        {"name": "watershed_area", "display_name": "Watershed Area", "type": FeatureSet},
        {"name": "snapped_points", "display_name": "Snapped Points", "type": FeatureSet}
    ]

    # use helper function to evaluate the input points and convert them, if necessary, to a FeatureSet
    input_fs = _evaluate_spatial_input(input_points)

    if input_fs.geometry_type != 'esriGeometryPoint':
        raise Exception('input_points FeatureSet must be point esriGeometryPoint, not {}.'.format(input_fs.geometry_type))

    input_fields = input_fs.fields
    if point_id_field and point_id_field not in [f['name'] for f in input_fields] and len(input_fields):
        input_fields_str = ','.join(input_fields)
        raise Exception('The provided point_id_field {} does not appear to be in the input_points FeatureSet fields - {}'.format(point_id_field, input_fields_str))

    if gis is None and arcgis.env.active_gis is None:
        raise Exception('GIS must be defined either by directly passing in a GIS object created using credentials, or one must already be created in the active Python session.')
    elif gis is None:
        gis = arcgis.env.active_gis

    url = gis.properties.helperServices.hydrology.url

    return _execute_gp_tool(gis, "Watershed", kwargs, param_db, return_values, True, url, future=future)



import warnings
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    import tensorflow as tf
    tf.enable_eager_execution()
    import onnx
    from onnx_tf.backend import prepare
import numpy as np
import math

def A(*a): return np.array(a[0]) if len(a)==1 else [np.array(o) for o in a]

imagenet_stats = A([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])

mean = 255* np.array(imagenet_stats[0], dtype=np.float32)
std  = 255* np.array(imagenet_stats[1], dtype=np.float32)

norm = lambda x: (x-mean)/ std
denorm = lambda x: x * std + mean

def pred2dict(bb_np, score, cat_str, c):
    # convert to top left x,y bottom right x,y
    return {"x1": bb_np[1],
            "x2": bb_np[3],
            "y1": bb_np[0],
            "y2": bb_np[2],
            "score": score,
            "category": cat_str,
            "class": c}

def to_np(x):
    return np.array(x)

def actn_to_bb(actn, anchors, grid_sizes):
        actn_bbs = tf.tanh(actn).numpy()
        actn_centers = (actn_bbs[:,:2]/2 * grid_sizes) + anchors[:,:2]
        actn_hw = (actn_bbs[:,2:]/2+1) * anchors[:,2:]
        return hw2corners(actn_centers, actn_hw)
    
def hw2corners(ctr, hw): 
        return tf.concat([ctr-hw/2, ctr+hw/2], axis=1).numpy()
    
def nms(boxes, scores, overlap=0.3, top_k=100):
    keep = []
    if len(boxes) == 0: return keep
    x1 = boxes[:, 0]
    y1 = boxes[:, 1]
    x2 = boxes[:, 2]
    y2 = boxes[:, 3]
    
    areas = (x2 - x1) * (y2 - y1)
    
    order = scores.argsort()
    order = order[-top_k:]
    
    while order.size > 0:
        i = order[-1]
        keep.append(i)
        order = order[:-1]
        xx1 = np.maximum(x1[i], x1[order])
        yy1 = np.maximum(y1[i], y1[order])
        xx2 = np.minimum(x2[i], x2[order])
        yy2 = np.minimum(y2[i], y2[order])
        w =  xx2 - xx1
        h =  yy2 - yy1
        inter = w * h
        rem_areas = areas[order]
        union = (rem_areas - inter) + areas[i]
        IoU = inter/union
        order = order[IoU<=overlap]

    return keep, len(keep)

def get_nms_preds(bb_class, bb_act, idx, anchors, grid_sizes, classes, nms_overlap, thres): 
    a_ic = actn_to_bb(bb_act[idx],anchors,grid_sizes)
    box_clas = tf.math.sigmoid((bb_class[idx]).max(1)).numpy()
    conf_scores = tf.transpose(tf.math.sigmoid(bb_class[idx])).numpy()

    out_scores, out_box, out_clas = [] ,[], []

    for i in range(1, len(conf_scores)):
        mask = conf_scores[i] > thres
        if mask.sum() == 0:
            continue
        scores = conf_scores[i][mask]
        mask1 = np.expand_dims(mask,1)*np.ones(a_ic.shape)
        mask1 = mask1.astype('bool')
        boxes = a_ic[mask1].reshape(-1,4)
        idx, cnt = nms(boxes, scores, nms_overlap, 50)
        idx = idx[:cnt]
        out_scores.append(scores[idx])
        out_box.append(boxes[idx])
        out_clas.append([i]*cnt)

    if out_clas == []:
        out_clas =[[0]]
    
    if out_box == []:
        out_box = [[]]

    if out_scores == []:
        out_scores = [[]]
        
    return predictions(out_box[0], out_clas[0], out_scores[0] if out_scores[0] is not None else None, thres, classes)

def predictions(bbox, clas=None, prs=None, thresh=0.3, classes=None): 
    if len(classes) is None:
        raise Exception("Classes are None")
    else:
        classes = ['bg'] + classes
    bb = bbox
    if prs is None:  prs  = [None]*len(bb)
    if clas is None: clas = [None]*len(bb)
    predictions = []
    for i, (b, c, pr) in enumerate(zip(bb, clas, prs)):
        if((b[2]>0) and (pr is None or pr > thresh)):
            cat_str = classes[c]
            score = pr * 100
            bb_np = b.astype('float64')
            predictions.append(pred2dict(bb_np, score, cat_str, c))
    return predictions

def detect_objects_image_space(graph, tiles,  anchors, grid_sizes, classes, nms_overlap, thres):
    tile_height, tile_width = tiles.shape[2], tiles.shape[3]

    tiles = tiles.astype('float32')
    img_t = norm(tiles.transpose(0,2,3,1))
    img_t = img_t.transpose(0,3,1,2)

    output_dict = graph.run(img_t)
    
    clas, bbox = output_dict._0, output_dict._1
    
    preds = { }

    for batch_idx in range(bbox.shape[0]):
        preds[batch_idx] = get_nms_preds(clas, bbox, batch_idx, anchors, grid_sizes, classes, nms_overlap, thres)

    batch_size = bbox.shape[0]
    side = math.sqrt(batch_size)

    num_boxes = 0
    for batch_idx in range(batch_size):
        num_boxes = num_boxes + len(preds[batch_idx])

    bounding_boxes = np.empty(shape=(num_boxes, 4), dtype=np.float)
    scores = np.empty(shape=(num_boxes), dtype=np.float)
    classes = np.empty(shape=(num_boxes), dtype=np.uint8)

    idx = 0
    for batch_idx in range(batch_size):
        i, j = batch_idx//side, batch_idx%side

        for pred in preds[batch_idx]:
            bounding_boxes[idx, 0] = (pred['y1'] + i)*tile_height
            bounding_boxes[idx, 1] = (pred['x1'] + j)*tile_width
            bounding_boxes[idx, 2] = (pred['y2'] + i)*tile_height
            bounding_boxes[idx, 3] = (pred['x2'] + j)*tile_width
            scores[idx] = pred['score']
            classes[idx] = pred['class']
            idx = idx+1

    return bounding_boxes, scores, classes




import os
import logging
import datetime as _datetime
import platform
import configparser

_log = logging.getLogger(__name__)

###########################################################################
class ProfileManager(object):
    """
    Allows for the controls and management of the
    profiles stored on the local operating system.
    """
    _gis = None
    _os = None
    _cfg_file_path = None
    #######################################################################
    def __init__(self):
        self._os = platform.system()
        self._cfg_file_path = os.path.expanduser("~") + '/.arcgisprofile'
        self._cfg_exists = os.path.isfile(self._cfg_file_path)
    #----------------------------------------------------------------------
    def _config_is_in_new_format(self, config):
        """ Any version <= 1.3.0 of the API used a different config file
        formatting that, among other things, did not store the last time
        a profile was modified. Thus, if 'date_modified' is found in at least
        one profile, it is in the new format
        """
        return any([profile_data for profile_data in config.values() \
                    if "date_modified" in profile_data])
    #----------------------------------------------------------------------
    def _update_profile_data_in_config(self, config, profile, url=None,
                                       username=None, key_file=None,
                                       cert_file=None, client_id=None):
        """Updates the specific profile in the config object to include
        any of the user defined arguments. This will overwrite old values.
        ***USE THIS FUNCTION INSTEAD OF MANUALLY MODIFYING PROFILE DATA***
        """
        if url is not None:
            config[profile]["url"] = url
            self._add_timestamp_to_profile_data_in_config(config, profile)
        if username is not None:
            config[profile]["username"] = username
            self._add_timestamp_to_profile_data_in_config(config, profile)
        if key_file is not None:
            config[profile]["key_file"] = key_file
            self._add_timestamp_to_profile_data_in_config(config, profile)
        if cert_file is not None:
            config[profile]["cert_file"] = cert_file
            self._add_timestamp_to_profile_data_in_config(config, profile)
        if client_id is not None:
            config[profile]["client_id"] = client_id
            self._add_timestamp_to_profile_data_in_config(config, profile)
    #----------------------------------------------------------------------
    def _add_timestamp_to_profile_data_in_config(self, config, profile):
        """Sets the 'date_modified' field to this moment's datetime"""
        config[profile]["date_modified"] = str(_datetime.datetime.now())
    #----------------------------------------------------------------------
    def _write_config(self, config, cfg_file_path):
        """write the config object to the .arcgisprofile file"""
        with open(cfg_file_path, "w") as writer:
            config.write(writer)
    #----------------------------------------------------------------------
    def _securely_store_password(self, profile, password):
        """Securely stores the password in an O.S. specific store via the
        keyring package. Can be retrieved later with just the profile name.

        If keyring is not properly set up system-wide, raise a RuntimeError
        """
        import keyring
        if self._current_keyring_is_recommended():
            return keyring.set_password("arcgis_python_api_profile_passwords",
                                        profile,
                                        password)
        else:
            raise RuntimeError(self._get_keyring_failure_message())
    #----------------------------------------------------------------------
    def _securely_get_password(self, profile):
        """Securely gets the profile specific password stored via keyring

        If keyring is not properly set up system-wide OR if a password is not
        found through keyring, log the respective warning and return 'None'
        """
        import keyring
        if self._current_keyring_is_recommended():
            # password will be None if no password is found for the profile
            password = keyring.get_password(
                "arcgis_python_api_profile_passwords", profile)
        else:
            password = None
            _log.warn(self._get_keyring_failure_message())

        if password is None:
            _log.warn("Profile {0} does not have a password on file through "\
                      "keyring. If you are expecting this behavior (PKI or "\
                      "IWA authentication, entering password through "\
                      "run-time prompt, etc.), please ignore this message. "\
                      "If you would like to store your password in the {0} "\
                      "profile, run GIS(profile = '{0}', password = ...). "\
                      "See the API doc for more details. "\
                      "(http://bit.ly/2CK2wG8)".format(profile))
        return password
    #----------------------------------------------------------------------
    def _securely_delete_password(self, profile):
        """Securely deletes the profile specific password via keyring

        If keyring is not properly set up system-wide, log a warning
        """
        import keyring
        if self._current_keyring_is_recommended():
            return keyring.delete_password(
                "arcgis_python_api_profile_passwords", profile)
        else:
            _log.warn(self._get_keyring_failure_message())
            return False
    #----------------------------------------------------------------------
    def _current_keyring_is_recommended(self):
        """The keyring project recommends 4 secure keyring backends. The
        defaults on Windows/OSX should be the recommended backends, but Linux
        needs some system-wide software installed and configured to securely
        function. Return if the current keyring is a supported, properly
        configured backend
        """
        import keyring
        supported_keyrings = [ keyring.backends.OS_X.Keyring,
                               keyring.backends.SecretService.Keyring,
                               keyring.backends.Windows.WinVaultKeyring,
                               keyring.backends.kwallet.DBusKeyring,
                               keyring.backends.chainer.ChainerBackend ]
        current_keyring = type(keyring.get_keyring())
        return current_keyring in supported_keyrings
    #----------------------------------------------------------------------
    def _get_keyring_failure_message(self):
        """An informative failure msg about the backend keyring being used"""
        import keyring
        return "Keyring backend being used ({}) either failed to install "\
               "or is not recommended by the keyring project (i.e. it is "\
               "not secure). This means you can not use stored passwords "\
               "through GIS's persistent profiles. Note that extra system-"\
               "wide steps must be taken on a Linux machine to use the python "\
               "keyring module securely. Read more about this at the "\
               "keyring API doc (http://bit.ly/2EWDP7B) and the ArcGIS API "\
               "for Python doc (http://bit.ly/2CK2wG8)."\
               "".format(keyring.get_keyring())
    #----------------------------------------------------------------------
    def list(self, as_df=False):
        """
        returns a list of profile names in the configuration file

        :returns: List if `as_df=False` or Pandas DataFrame if `as_df=True`
        """
        if self._cfg_exists and as_df == False:
            config = configparser.ConfigParser()
            config.read(self._cfg_file_path)
            return config.sections()
        elif self._cfg_exists and as_df:
            import pandas as pd
            all_profiles = []
            for p in self.list():
                all_profiles.append(self.get(p))
            return pd.DataFrame(data=all_profiles)
        return []
    #--------------------------------------------------------------------------
    def get(self, profile):
        """
        Returns the profile information for a given entry.

        =====================================================================     ====================================================================
        **Parameter**                                                             **Description**
        ---------------------------------------------------------------------     --------------------------------------------------------------------
        profile                                                                   Required String. The name of the profile to get the information about.
        =====================================================================     ====================================================================

        :returns: Dict

        """
        profile_file = self._cfg_file_path
        if self._cfg_exists:
            config = configparser.ConfigParser()
            config.read(profile_file)
            keys = config.options(profile)
            values = {}
            for key in keys:
                try:
                    if key == 'date_modified':

                        values[key] = _datetime.datetime.strptime(
                            config.get(
                                profile,
                                key),
                            "%Y-%m-%d %H:%M:%S.%f")
                    else:
                        values[key] = config.get(profile, key)
                except:
                    values[key] = None
            return values
        return None
    #--------------------------------------------------------------------------
    def delete(self, profile):
        """
        Deletes a profile from the .arcgisprofile file

        =====================================================================     ====================================================================
        **Parameter**                                                             **Description**
        ---------------------------------------------------------------------     --------------------------------------------------------------------
        profile                                                                   Required String. The name of the profile to delete.
        =====================================================================     ====================================================================

        :returns: Boolean
        """
        import keyring
        profile_file = self._cfg_file_path
        if self._cfg_exists:
            profiles = self.list()
            if profile in profiles:
                config = configparser.ConfigParser()
                with open(profile_file, 'r') as reader:
                    config.read_file(reader)
                data = dict(config.items(profile))
                try:
                    keyring.delete_password(service_name="arcgis_python_api_profile_passwords",
                                            username=profile)
                except: pass
                config.remove_section(section=profile)
                with open(profile_file, "w") as f:
                    config.write(f)

                return True
            else:
                raise ValueError("Profile not found.")
        return False
    #----------------------------------------------------------------------
    def update(self, profile, url=None,
               username=None, password=None,
               key_file=None, cert_file=None,
               client_id=None):
        """
        Updates an existing profile in the credential manager.

        =====================================================================     ====================================================================
        **Parameter**                                                             **Description**
        ---------------------------------------------------------------------     --------------------------------------------------------------------
        profile                                                                   Required String. The name of the profile to update.
        ---------------------------------------------------------------------     --------------------------------------------------------------------
        url                                                                       Optional String.  The site url.
        ---------------------------------------------------------------------     --------------------------------------------------------------------
        username                                                                  Optional String.  The login user name.
        ---------------------------------------------------------------------     --------------------------------------------------------------------
        password                                                                  Optional String.  The login user password.
        ---------------------------------------------------------------------     --------------------------------------------------------------------
        key_file                                                                  Optional String.  The key file for PKI security.
        ---------------------------------------------------------------------     --------------------------------------------------------------------
        cert_file                                                                 Optional String.  The cert file for PKI security.
        ---------------------------------------------------------------------     --------------------------------------------------------------------
        client_id                                                                 Optional String.  The client ID for oauth login.
        =====================================================================     ====================================================================

        :returns: boolean

        """
        if profile not in self.list():
            raise ValueError("Could not find profile {}. Use `create` to generate a new profile.".format(profile))
        _log.info("Updating profile {} ...".format(profile))
        return self.create(profile, url, username, password, key_file, cert_file, client_id)
    #----------------------------------------------------------------------
    def create(self, profile, url=None,
               username=None, password=None,
               key_file=None, cert_file=None,
               client_id=None):
        """
        Adds a new entry into the Profile Store.

        =====================================================================     ====================================================================
        **Parameter**                                                             **Description**
        ---------------------------------------------------------------------     --------------------------------------------------------------------
        profile                                                                   Required String. The name of the profile to add.
        ---------------------------------------------------------------------     --------------------------------------------------------------------
        url                                                                       Optional String.  The site url.
        ---------------------------------------------------------------------     --------------------------------------------------------------------
        username                                                                  Optional String.  The login user name.
        ---------------------------------------------------------------------     --------------------------------------------------------------------
        password                                                                  Optional String.  The login user password.
        ---------------------------------------------------------------------     --------------------------------------------------------------------
        key_file                                                                  Optional String.  The key file for PKI security.
        ---------------------------------------------------------------------     --------------------------------------------------------------------
        cert_file                                                                 Optional String.  The cert file for PKI security.
        ---------------------------------------------------------------------     --------------------------------------------------------------------
        client_id                                                                 Optional String.  The client ID for oauth login.
        =====================================================================     ====================================================================

        :returns: boolean

        """
        try:
            config = configparser.ConfigParser()
            if self._cfg_exists:
                config.read(self._cfg_file_path)
            if profile not in self.list():
                _log.info("Adding new profile {} to config...".format(profile))
                config.add_section(profile)
                self._add_timestamp_to_profile_data_in_config(config, profile)
            self._update_profile_data_in_config(config=config,
                                                profile=profile,
                                                url=url,
                                                username=username,
                                                key_file=key_file,
                                                cert_file=cert_file,
                                                client_id=client_id)
            if password is not None:
                self._securely_store_password(profile, password)
            self._write_config(config, self._cfg_file_path)
            self._cfg_exists = True
            return True
        except:
            return False
    #----------------------------------------------------------------------
    def save_as(self, profile, gis):
        """
        Saves and adds the provided `GIS` to the profile.

        =====================================================================     ====================================================================
        **Parameter**                                                             **Description**
        ---------------------------------------------------------------------     --------------------------------------------------------------------
        name                                                                      Required String. The name of the profile to save.
        ---------------------------------------------------------------------     --------------------------------------------------------------------
        gis                                                                       Required GIS. The connection object to update the profile with.
        =====================================================================     ====================================================================

        :returns: Boolean

        """
        from arcgis.gis import GIS
        url = gis._url
        u = gis._username
        p = gis._password
        kf = gis._key_file
        cf = gis._cert_file
        ci = gis._client_id
        if profile.lower() in [p.lower() for p in self.list()]:
            cfg_file_path = self._cfg_file_path
            config = configparser.ConfigParser()
            if os.path.isfile(cfg_file_path):
                config.read(cfg_file_path)
            self._update_profile_data_in_config(config, profile=profile, url=url, username=u,
                                               key_file=kf, cert_file=cf, client_id=ci)
            if p is not None:
                self._securely_store_password(profile, p)
            self._write_config(config, cfg_file_path)
            return True
        else:
            return self.create(config, profile=profile, url=url, username=u,
                        key_file=kf, cert_file=cf, client_id=ci)
        return False
    #----------------------------------------------------------------------
    def _retrieve(self, profile):
        """gets the login information"""
        url, username, password, key_file, cert_file, client_id = None, None, None, None, None, None
        if profile.lower() in [p.lower() for p in self.list()]:
            cfg_file_path = self._cfg_file_path
            config = configparser.ConfigParser()
            if os.path.isfile(cfg_file_path):
                config.read(cfg_file_path)
            if config.has_option(profile,   "url"):
                url = config[profile]["url"]
            if config.has_option(profile,   "username"):
                username =  config[profile]["username"]
            if config.has_option(profile,   "key_file"):
                key_file =  config[profile]["key_file"]
            if config.has_option(profile,   "cert_file"):
                cert_file = config[profile]["cert_file"]
            if config.has_option(profile,   "client_id"):
                client_id = config[profile]["client_id"]

            password = self._securely_get_password(profile)
        return url, username, password, key_file, cert_file, client_id



"""
The arcgis.geoanalytics module provides types and functions for distributed analysis of large datasets.
These GeoAnalytics tools work with big data registered in the GISs datastores as well as with feature layers.

Use arcgis.geoanalytics.is_analysis_supported(gis) to check if geoanalytics is supported in your GIS.

Note: GeoAnalytics operations use the following context parameters defined in the `arcgis.env` module:

        =========================     ====================================================================
        **Context parameter**         **Description**
        -------------------------     --------------------------------------------------------------------
        out_spatial_reference         Used for setting the output spatial reference
        -------------------------     --------------------------------------------------------------------
        process_spatial_reference     Used for setting the processing spatial reference.
        -------------------------     --------------------------------------------------------------------
        analysis_extent               Used for setting the analysis extent.
        -------------------------     --------------------------------------------------------------------
        output_datastore              Used for setting the output datastore to be used.
        =========================     ====================================================================
"""

from . import summarize_data, analyze_patterns, use_proximity, manage_data, find_locations, data_enrichment

def get_datastores(gis=None):
    """
    Returns a helper object to manage geoanalytics datastores in the GIS.
    If a gis isn't specified, returns datastore manager of arcgis.env.active_gis
    """
    import arcgis
    gis = arcgis.env.active_gis if gis is None else gis

    for ds in gis._datastores:
        if 'GeoAnalytics' in ds._server['serverFunction']:
            return ds

    return None

def define_output_datastore(datastore=None, template=None):
    """
    Sets the `arcgis.env.output_datastore` by providing the datastore and template name
    to this method. If datastore is None, the `arcgis.env.output_datastore` will reset
    to default.

    ==========================   ===============================================================
    **Argument**                 **Description**
    --------------------------   ---------------------------------------------------------------
    datastore                    Optional Datastore/String. This specifies the big data file
                                 share to save GeoAnalyticss results to. If specified as None the
                                 `arcgis.env.output_datastore` will reset to default.  Allowed
                                 string values are: `spatiotemporal` or `relational`.
    --------------------------   ---------------------------------------------------------------
    template                     Optional string. When specified, the `template` determines how
                                 GeoAnalytics result schema will be formatted. The output will
                                 be written to a file in the big data file share.
    ==========================   ===============================================================

    :returns: Boolean

    """
    import arcgis
    if isinstance(datastore, str) and \
       str(datastore).lower() in ["spatiotemporal", "relational"]:
        arcgis.env.output_datastore = str(datastore).lower()
        return True
    elif isinstance(datastore, str) and \
       not str(datastore).lower() in ["spatiotemporal", "relational"]:
        raise ValueError("datastore can only ")
    elif datastore and template:
        if isinstance(datastore, arcgis.gis.Datastore):
            arcgis.env.output_datastore = "{path}:{template}".format(path=datastore.path,
                                                                     template=template)
        elif isinstance(datastore, str):
            arcgis.env.output_datastore = "{path}:{template}".format(path=datastore,
                                                                     template=template)
        elif isinstance(datastore, arcgis.gis.server.Datastore):
            path = datastore.properties.path
            arcgis.env.output_datastore = "{path}:{template}".format(path=datastore,
                                                                     template=template)
        else:
            raise ValueError("Invalid Datastore")
        return True
    elif template and datastore is None:
        raise ValueError("datastore must be specified to set an output template.")
    elif datastore and template is None:
        raise ValueError("template must be specified to set the output_datastore.")
    else:
        arcgis.env.output_datastore = None
        return True
    return False

def is_supported(gis=None):
    """
    Returns True if the GIS supports geoanalytics. If a gis isn't specified,
    checks if arcgis.env.active_gis supports geoanalytics
    """
    import arcgis
    gis = arcgis.env.active_gis if gis is None else gis
    if 'geoanalytics' in gis.properties.helperServices:
        return True
    else:
        return False




from fastai.vision import ImageSegment, Image
from fastai.vision.image import open_image, show_image, pil2tensor
from fastai.vision.data import SegmentationProcessor, ImageList
from fastai.layers import CrossEntropyFlat
from fastai.callbacks import LearnerCallback
import torch
import warnings
import PIL
import numpy as np
from skimage import io
import matplotlib.pyplot as plt

class ArcGISImageSegment(Image):
    "Support applying transforms to segmentation masks data in `px`."
    def __init__(self, x, cmap=None, norm=None):
        super(ArcGISImageSegment, self).__init__(x)
        self.cmap = cmap
        self.mplnorm = norm
        self.type = np.unique(x)

    def lighting(self, func, *args, **kwargs):
        return self

    def refresh(self):
        self.sample_kwargs['mode'] = 'nearest'
        return super().refresh()

    @property
    def data(self):
        "Return this image pixels as a `LongTensor`."
        return self.px.long()

    def show(self, ax = None, figsize = (3,3), title = None, hide_axis = True, cmap='tab20', alpha = 0.5, **kwargs):

        if ax is None: fig,ax = plt.subplots(figsize=figsize)
        masks = self.data[0].numpy()
        for i in range(1, self.data.shape[0]):
            max_unique = np.max(np.unique(masks))
            mask = np.where(self.data[i]>0, self.data[i] + max_unique, self.data[i])
            masks += mask
        ax.imshow(masks, cmap=cmap, alpha=alpha, **kwargs)
        if hide_axis: ax.axis('off')
        if title: ax.set_title(title)


def is_no_color(color_mapping):
    if isinstance(color_mapping, dict):
        color_mapping = list(color_mapping.values())
    return (np.array(color_mapping) == [-1., -1., -1.]).any()

class ArcGISSegmentationLabelList(ImageList):
    "`ItemList` for segmentation masks."
    _processor = SegmentationProcessor
    def __init__(self, items, chip_size, classes=None, class_mapping=None, color_mapping=None, **kwargs):
        super().__init__(items, **kwargs)
        self.class_mapping = class_mapping
        self.color_mapping = color_mapping
        self.copy_new.append('classes')
        self.classes, self.loss_func = classes, CrossEntropyFlat(axis=1)
        self.chip_size = chip_size
        self.inverse_class_mapping = {}
        for k, v in self.class_mapping.items():
            self.inverse_class_mapping[v] = k
        if is_no_color(list(color_mapping.values())):
            self.cmap = 'tab20'  ## compute cmap from palette
            import matplotlib as mpl
            bounds = list(color_mapping.keys())
            if len(bounds) < 3: # Two handle two classes i am adding one number to the classes which is not already in bounds
                bounds = bounds + [max(bounds)+1]
            self.mplnorm = mpl.colors.BoundaryNorm(bounds, len(bounds))
        else:
            import matplotlib as mpl
            bounds = list(color_mapping.keys())
            if len(bounds) < 3: # Two handle two classes i am adding one number to the classes which is not already in bounds
                bounds = bounds + [max(bounds)+1]
            self.cmap = mpl.colors.ListedColormap(np.array(list(color_mapping.values()))/255)
            self.mplnorm = mpl.colors.BoundaryNorm(bounds, self.cmap.N)

        if len(color_mapping.keys()) == 1:
            self.cmap = 'tab20'
            self.mplnorm = None

    def open(self, fn):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", UserWarning) # EXIF warning from TiffPlugin
            if len(fn) != 0:
                img_shape = io.imread(fn[0]).shape
            else:
                labeled_mask = torch.zeros((len(self.class_mapping), self.chip_size, self.chip_size))
                return ArcGISImageSegment(labeled_mask, cmap=self.cmap, norm=self.mplnorm)
            k = 0

            labeled_mask = np.zeros((1, img_shape[0], img_shape[1]))

            for j in range(len(self.class_mapping)):

                if k < len(fn):
                    lbl_name = int(self.inverse_class_mapping[fn[k].parent.name])
                else:
                    lbl_name = len(self.class_mapping) + 2
                if lbl_name == j+1:                    
                    img = io.imread(fn[k])
                    k = k + 1
                    if len(img.shape)==3:
                        img = img.transpose(2,0,1)
                        img_mask = img[0]
                        for i in range(1, img.shape[0]):
                            max_unique = np.max(np.unique(img_mask))
                            img_i = np.where(img[i]>0, img[i] + max_unique, img[i])
                            img_mask += img_i
                        img_mask = np.expand_dims(img_mask, axis = 0)
                    else:
                        img_mask = np.expand_dims(img, axis = 0)
                else:
                    img_mask = np.zeros((1, img_shape[0], img_shape[1]))
                labeled_mask = np.append(labeled_mask, img_mask, axis = 0)
            labeled_mask = labeled_mask[1:,:,:]
            labeled_mask = torch.Tensor(list(labeled_mask))
        return ArcGISImageSegment(labeled_mask, cmap=self.cmap, norm=self.mplnorm)

    def reconstruct(self, t): 
        return ArcGISImageSegment(t, cmap=self.cmap, norm=self.mplnorm)

class ArcGISInstanceSegmentationItemList(ImageList):
    "`ItemList` suitable for segmentation tasks."
    _label_cls, _square_show_res = ArcGISSegmentationLabelList, False

def mask_rcnn_loss(loss_value, *args):

    final_loss = 0.
    for i in loss_value.values():
        if not (torch.isnan(i) or torch.isinf(i)):
            final_loss += i
            
    return final_loss

class train_callback(LearnerCallback):

    def __init__(self, learn):
        super().__init__(learn)
   
    def on_batch_begin(self, last_input, last_target, **kwargs):
        "Handle new batch `xb`,`yb` in `train` or validation."

        target_list = []
        for i in range(len(last_target)):

            boxes =  []
            masks = np.zeros((1, last_target[i].shape[1], last_target[i].shape[2]))
            labels = []
            for j in range(last_target[i].shape[0]):

                mask = np.array(last_target[i].data[j])
                obj_ids = np.unique(mask)

                if len(obj_ids)==1:
                    continue

                obj_ids = obj_ids[1:]
                mask_j = mask == obj_ids[:, None, None]
                num_objs = len(obj_ids)
                for k in range(num_objs):
                    pos = np.where(mask_j[k])
                    xmin = np.min(pos[1])
                    xmax = np.max(pos[1])
                    ymin = np.min(pos[0])
                    ymax = np.max(pos[0])
                    boxes.append([xmin, ymin, xmax, ymax])

                masks = np.append(masks, mask_j, axis = 0)
                labels_j = torch.ones((num_objs,), dtype=torch.int64)
                labels_j = labels_j*(j+1)
                labels.append(labels_j)
            
            if(masks.shape[0]==1): # if no object in image
                masks[0,50:51,50:51] = 1
                labels = torch.tensor([0])
                boxes = torch.tensor([[50.,50.,51.,51.]])
            else:
                labels = torch.cat(labels)
                boxes = torch.as_tensor(boxes, dtype=torch.float32)
                masks = masks[1:,:,:]
            masks = torch.as_tensor(masks, dtype=torch.uint8)
            target = {}
            target["boxes"] = boxes.cuda()
            target["labels"] = labels.cuda()
            target["masks"] = masks.cuda()
            target_list.append(target)

        self.learn.model.train()
        last_input = [last_input, target_list]
        last_target = [torch.tensor([1]) for i in last_target]
        return {'last_input':last_input, 'last_target':last_target}            


import arcgis
from pathlib import Path
import os
import time
import tempfile
import json
import logging
from .._data import _raise_fastai_import_error
from warnings import warn
import contextlib
import io
import sys
import socket

HAS_FASTAI = True
HAS_TENSORBOARDX = True

try:
    from fastai.callbacks import TrackerCallback, EarlyStoppingCallback, LearnerCallback
    from torch import nn
    import torch
    from torchvision import models
    import math
    import warnings
except ImportError:
    HAS_FASTAI = False
    class TrackerCallback():
        pass
    class LearnerCallback():
        pass

try:
    import tensorboardX 
    # LearnerTensorboardWriter uses SummaryWriter from tensorboardX
    from fastai.callbacks.tensorboard import LearnerTensorboardWriter
except:
    HAS_TENSORBOARDX = False

logger = logging.getLogger()

#For lr computation, skip beginning and trailing values.
losses_skipped = 5
trailing_losses_skipped = 5
model_characteristics_folder = 'ModelCharacteristics'


@contextlib.contextmanager
def nostdout():
    save_stdout = sys.stdout
    sys.stdout = io.BytesIO()
    yield
    sys.stdout = save_stdout
    
class _EmptyData():
    def __init__(self, path, c, loss_func, chip_size):
        self.path = path
        if getattr(arcgis.env, "_processorType", "") == "GPU" and torch.cuda.is_available():
            self.device = torch.device("cuda")
        elif getattr(arcgis.env, "_processorType", "") == "CPU":
            self.device = torch.device("cpu")
        else:
            self.device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')
        self.c = c
        self.loss_func = loss_func
        self.chip_size = chip_size


class _MultiGPUCallback(LearnerCallback):
    """
    Parallize over multiple GPUs only if multiple GPUs are present.
    """
    def __init__(self, learn):
        super(_MultiGPUCallback, self).__init__(learn)
        
        self.multi_gpu = torch.cuda.device_count() > 1

    def on_train_begin(self, **kwargs):
        if self.multi_gpu:
            logger.info('Training on multiple GPUs')
            self.learn.model = nn.DataParallel(self.learn.model)
    
    def on_train_end(self, **kwargs):
        if self.multi_gpu:
            self.learn.model = self.learn.model.module

def _set_multigpu_callback(model):
    model.learn.callback_fns.append(_MultiGPUCallback)


def _create_zip(zipname, path):
    import shutil
    if os.path.exists(os.path.join(path, zipname) + '.dlpk'):
        os.remove(os.path.join(path, zipname) + '.dlpk')
        
    temp_dir = tempfile.TemporaryDirectory().name    
    zip_file = shutil.make_archive(os.path.join(temp_dir, zipname), 'zip', path)
    dlpk_base = os.path.splitext(zip_file)[0]
    os.rename(zip_file, dlpk_base + '.dlpk')
    dlpk_file = dlpk_base+'.dlpk'
    shutil.move(dlpk_file, path)


class SaveModelCallback(TrackerCallback):

    def __init__(self, model, every='improvement', name='bestmodel', load_best_at_end=True, **kwargs):
        super().__init__(learn=model.learn, **kwargs)
        self.model = model
        self.every = every
        self.name = name
        self.load_best_at_end = load_best_at_end
        if self.every not in ['improvement', 'epoch']:
            warn('SaveModel every {} is invalid, falling back to "improvement".'.format(self.every))
            self.every = 'improvement'

    def on_epoch_end(self, epoch, **kwargs):
        "Compare the value monitored to its best score and maybe save the model."
        if self.every == "epoch": self.model.save('{}_{}'.format(self.name, epoch))
        else: #every="improvement"
            current = self.get_monitor_value()
            if current is not None and self.operator(current, self.best):
                if arcgis.env.verbose:
                    print('saving checkpoint.')
                self.best = current
                self.model._save('{}'.format(self.name), zip_files=False, save_html=False)

    def on_train_end(self, **kwargs):
        "Load the best model."      
        if self.every == "improvement" and self.load_best_at_end:
            try:
                self.model.load('{}'.format(self.name))
            except FileNotFoundError:
                pass
            
            try:
                self.model.save('{}'.format(self.name))
            except:
                pass

def _get_tail(model):
    index_order = 0
    first_layer = None
    try:
        first_layer = model._modules[list(model._modules.keys())[0]]
        while True:
            first_layer = first_layer[0]
            index_order+=1
    except:
        pass
    return first_layer, index_order

def _get_ms_tail(tail, data, type_init='average'):
    new_tail = tail.__class__(
        in_channels=len(data._extract_bands), 
        out_channels=tail.out_channels,
        kernel_size=tail.kernel_size,
        stride=tail.stride,
        padding=tail.padding,
        dilation=tail.dilation,
        groups=tail.groups,
        bias=tail.bias is not None,
        padding_mode=tail.padding_mode,
    )
    if type_init == 'average':
        rgb_weights = tail.weight.data
        avg_weights = tail.weight.data.mean(dim=1)
        rgb_map = {'r':0, 'g':1, 'b': 2}
        for i, j in enumerate(data._extract_bands):
            band = str(data._bands[j]).lower()
            b = rgb_map.get(band, None)
            #print(b)
            if b is not None:
                new_tail.weight.data[:, i] = tail.weight.data[:, b]
            else:
                #print('unknown band')
                new_tail.weight.data[:, i] = tail.weight.data[:, 0] # Red Band Wieghts for all other band weights
    return new_tail

def _set_tail(model, new_tail, index_order=0, inplace=True):
    i = 0
    codeblock = 'model._modules[list(model._modules.keys())[0]]'
    while i < index_order:
        codeblock+='[0]'
        i+=1
    exec(codeblock + ' = new_tail')
    
    #first_layer = model._modules[list(model._modules.keys())[0]]
    #i = 0
    #while i < index_order:
    #    first_layer = first_layer[0]
    #    i+=1
    #first_layer = new_tail
    
    if not inplace:
        return model

def _change_tail(model, bands):
        tail, index_order = _get_tail(model)
        new_tail = _get_ms_tail(tail, bands)
        _set_tail(
            model, 
            new_tail, 
            index_order,
            inplace=True
        )
        return model

class ArcGISModel(object):
    
    def __init__(self, data, backbone=None, **kwargs):
        if not HAS_FASTAI:
            _raise_fastai_import_error()

        if getattr(arcgis.env, "_processorType", "") == "GPU" and torch.cuda.is_available():
            self._device = torch.device("cuda")
        elif getattr(arcgis.env, "_processorType", "") == "CPU":
            self._device = torch.device("cpu")
        else:
            self._device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')

        if backbone is None:
            self._backbone = models.resnet34
        elif type(backbone) is str:
            if hasattr(models, backbone):
                self._backbone = getattr(models, backbone)
            elif hasattr(models.detection, backbone):
                self._backbone = getattr(models.detection, backbone)
        else:
            self._backbone = backbone

        if hasattr(data, '_is_multispectral'): # multispectral support
            self._is_multispectral = getattr(data, '_is_multispectral')
        else:
            self._is_multispectral = False
        if self._is_multispectral:
            self._imagery_type = data._imagery_type   
            self._bands = data._bands
            self._backbone_ = self._backbone
            def backbone_wrapper(pretrained):
                return _change_tail(self._backbone_(pretrained), data)
            self._backbone = backbone_wrapper

        self.learn = None
        self._data = data
        self._learning_rate = None

        # Declare the family of backbones to be unpacked and used by different models as supported types
        self._vgg_family = [models.vgg11.__name__, models.vgg11_bn.__name__, models.vgg13.__name__, models.vgg13_bn.__name__, 
                            models.vgg16.__name__, models.vgg16_bn.__name__, models.vgg19.__name__, models.vgg19_bn.__name__]
        self._resnet_family = [models.resnet18.__name__, models.resnet34.__name__, models.resnet50.__name__, 
                               models.resnet101.__name__, models.resnet152.__name__]
        self._densenet_family = [models.densenet121.__name__, models.densenet169.__name__, models.densenet161.__name__, 
                                 models.densenet201.__name__]

    def _check_backbone_support(self, backbone):
        "Fetches the backbone name and returns True if it is in the list of supported backbones"
        backbone_name = backbone if type(backbone) is str else backbone.__name__
        return False if backbone_name not in self.supported_backbones else True
    
    def _arcgis_init_callback(self):
        if self._is_multispectral:
            if self._data._train_tail:
                next(self.learn.model.parameters()).requires_grad = True # make first conv weights learnable
                self.learn.create_opt(slice(3e-3))
            if hasattr(self, '_show_results_multispectral'):
                self.show_results = self._show_results_multispectral
            
    def lr_find(self, allow_plot=True):
        """
        Runs the Learning Rate Finder, and displays the graph of it's output.
        Helps in choosing the optimum learning rate for training the model.
        """
        self.learn.lr_find()
        from IPython.display import clear_output
        clear_output()
        lr, index = self._find_lr()
        if allow_plot:
            self._show_lr_plot(index)

        return lr

    def _show_lr_plot(self, index):
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots(1, 1)
        ax.plot(
            self.learn.recorder.lrs[losses_skipped:-trailing_losses_skipped],
            self.learn.recorder.losses[losses_skipped:-trailing_losses_skipped]
        )
        ax.set_ylabel("Loss")
        ax.set_xlabel("Learning Rate")
        ax.set_xscale('log')
        ax.xaxis.set_major_formatter(plt.FormatStrFormatter('%.0e'))
        ax.plot(
            self.learn.recorder.lrs[index],
            self.learn.recorder.losses[index],
            markersize=10,
            marker='o',
            color='red'
        )

        plt.show()

    def _find_lr(self):
        losses = self.learn.recorder.losses[losses_skipped:-trailing_losses_skipped]
        lrs = self.learn.recorder.lrs[losses_skipped:-trailing_losses_skipped]

        n = len(losses)

        max_start = 0
        max_end = 0

        lds = [1] * n

        for i in range(1, n):
            for j in range(0, i):
                if losses[i] < losses[j] and lds[i] < lds[j] + 1:
                    lds[i] = lds[j] + 1
                if lds[max_end] < lds[i]:
                    max_end = i
                    max_start = max_end - lds[max_end]

        sections = (max_end - max_start) / 3
        final_index = max_start + int(sections) + int(sections/2)

        return lrs[final_index], losses_skipped + final_index

    @property
    def _model_metrics(self):
        raise NotImplementedError

    def fit(self, epochs=10, lr=None, one_cycle=True, early_stopping=False, checkpoint=True, tensorboard=False, **kwargs):
        """
        Train the model for the specified number of epochs and using the
        specified learning rates
        
        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        epochs                  Required integer. Number of cycles of training
                                on the data. Increase it if underfitting.
        ---------------------   -------------------------------------------
        lr                      Optional float or slice of floats. Learning rate
                                to be used for training the model. If ``lr=None``, 
                                an optimal learning rate is automatically deduced 
                                for training the model.
        ---------------------   -------------------------------------------
        one_cycle               Optional boolean. Parameter to select 1cycle
                                learning rate schedule. If set to `False` no 
                                learning rate schedule is used.       
        ---------------------   -------------------------------------------
        early_stopping          Optional boolean. Parameter to add early stopping.
                                If set to 'True' training will stop if validation
                                loss stops improving for 5 epochs.        
        ---------------------   -------------------------------------------
        checkpoint              Optional boolean. Parameter to save the best model
                                during training. If set to `True` the best model 
                                based on validation loss will be saved during 
                                training.
        ---------------------   -------------------------------------------
        tensorboard             Optional boolean. Parameter to write the training log. 
                                If set to 'True' the log will be saved at 
                                <dataset-path>/training_log which can be visualized in
                                tensorboard. Required tensorboardx version=1.7 (Experimental support).

                                The default value is 'False'.
        =====================   ===========================================
        """
        if lr is None:
            print('Finding optimum learning rate.')

            lr = self.lr_find(allow_plot=False)
            lr = slice(lr/10, lr)

        self._learning_rate = lr

        if arcgis.env.verbose:
            logger.info('Fitting the model.')        
        
        callbacks = kwargs['callbacks'] if 'callbacks' in kwargs.keys() else []
        kwargs.pop('callbacks', None)
        if early_stopping:
            callbacks.append(EarlyStoppingCallback(learn=self.learn, monitor='valid_loss', min_delta=0.01, patience=5))
        if checkpoint:
            from datetime import datetime
            now = datetime.now()
            callbacks.append(SaveModelCallback(self, monitor='valid_loss', every='improvement', name=now.strftime("checkpoint_%Y-%m-%d_%H-%M-%S")))
        
        # If tensorboardx is installed write a log with name as timestamp
        if tensorboard and HAS_TENSORBOARDX:
            training_id = time.strftime("log_%Y-%m-%d_%H-%M-%S")
            log_path = Path(os.path.dirname(self._data.path)) / 'training_log'
            callbacks.append(LearnerTensorboardWriter(learn=self.learn, base_dir=log_path, name=training_id))
            hostname = socket.gethostname()
            print("Monitor training using Tensorboard using the following command: 'tensorboard --host={} --logdir={}'".format(hostname, log_path))
        # Send out a warning if tensorboardX is not installed
        elif tensorboard:
            warn("Install tensorboardX 1.7 'pip install tensorboardx==1.7' to write training log")

        if one_cycle:
            self.learn.fit_one_cycle(epochs, lr, callbacks=callbacks, **kwargs)
        else:
            self.learn.fit(epochs, lr, callbacks=callbacks, **kwargs)
        
    def unfreeze(self):
        """
        Unfreezes the earlier layers of the model for fine-tuning.
        """
        self.learn.unfreeze()

    def _create_emd(self, path):
        backbone = self._backbone.__name__
        if backbone == 'backbone_wrapper':
            backbone = self._backbone_.__name__
        self._emd_template = {
            'ModelFile': path.name,
            'ImageHeight': self._data.chip_size,
            'ImageWidth': self._data.chip_size,
            'ModelParameters': {'backbone': backbone},
            'LearningRate': str(self._learning_rate),
            'ModelName': type(self).__name__
        }

        model_metrics = self._model_metrics

        if model_metrics.get('accuracy'):
            self._emd_template['accuracy'] = model_metrics.get('accuracy')
        
        if model_metrics.get('average_precision_score'):
            self._emd_template['average_precision_score'] = model_metrics.get('average_precision_score')

        resize_to = None
        if hasattr(self._data, 'resize_to') and self._data.resize_to:
            resize_to = self._data.resize_to

        self._emd_template['resize_to'] = resize_to

    @staticmethod
    def _create_html(path_model):
        import base64

        model_characteristics_dir = os.path.join(path_model.parent.absolute(), model_characteristics_folder)
        loss_graph = os.path.join(model_characteristics_dir, 'loss_graph.png')
        show_results = os.path.join(model_characteristics_dir, 'show_results.png')
        confusion_matrix = os.path.join(model_characteristics_dir, 'confusion_matrix.png')

        encoded_losses_img = None
        if os.path.exists(loss_graph):
            encoded_losses_img = "data:image/png;base64,{0}".format(base64.b64encode(open(loss_graph, 'rb').read()).decode('utf-8'))

        encoded_showresults = None
        if os.path.exists(show_results):
            encoded_showresults = "data:image/png;base64,{0}".format(base64.b64encode(open(show_results, 'rb').read()).decode('utf-8'))

        confusion_matrix_img = None
        if os.path.exists(confusion_matrix):
            confusion_matrix_img = "data:image/png;base64,{0}".format(base64.b64encode(open(confusion_matrix, 'rb').read()).decode('utf-8'))

        html_file_path = os.path.join(path_model.parent, 'model_metrics.html')

        emd_path = os.path.join(path_model.parent, path_model.stem + '.emd')
        if not os.path.exists(emd_path):
            return

        emd_template = json.load(open(emd_path, 'r'))

        HTML_TEMPLATE = f"""        
                <p><b> {emd_template.get("ModelName").replace('>', '').replace('<', '')} </b></p>
                <p><b>Backbone:</b> {emd_template.get('ModelParameters', {}).get('backbone')}</p>
                <p><b>Learning Rate:</b> {emd_template.get('LearningRate')}</p>
                <p><b>Training and Validation loss</b></p>
                <img src="{encoded_losses_img}" alt="training and validation losses">
        """

        model_analysis = None
        if confusion_matrix_img:
             model_analysis = f""" <p><b>Confusion Matrix</p></b>
                    <img src="{confusion_matrix_img}" alt="Confusion Matrix" width="500" height="333">
            """
        if emd_template.get('accuracy'):
            model_analysis = f"""
            <p><b>Accuracy:</b> {emd_template.get('accuracy')}</p>
        """

        if emd_template.get('average_precision_score'):
            model_analysis = f"""
            <p><b>Average Precision Score:</b> {emd_template.get('average_precision_score')}</p>
        """

        if model_analysis:
            HTML_TEMPLATE += f"""
            <p><b>Analysis of the model</b></p>
            {model_analysis}
        """

        HTML_TEMPLATE += f"""
            <p><b>Sample Results</b></p>
            <img src="{encoded_showresults}" alt="Sample Results">
        """

        file = open(html_file_path, 'w')
        file.write(HTML_TEMPLATE)
        file.close()

    def _save(self, name_or_path, framework='PyTorch', zip_files=True, save_html=True, publish=False, gis=None, **kwargs):
        temp = self.learn.path

        if '\\' in name_or_path or '/' in name_or_path:
            path = Path(name_or_path)
            name = path.parts[-1]
            # to make fastai save to both path and with name    
            self.learn.path = path
            self.learn.model_dir = ''
            if not os.path.exists(self.learn.path):
                os.makedirs(self.learn.path)
        else:
            # fixing fastai bug
            self.learn.path = self.learn.path.parent
            self.learn.model_dir =  Path(self.learn.model_dir) /  name_or_path
            if not os.path.exists(self.learn.path / self.learn.model_dir):
                os.makedirs(self.learn.path / self.learn.model_dir)
            name = name_or_path

        try:
            saved_path = self.learn.save(name,  return_path=True)
            # undoing changes to self.learn.path
        except Exception as e:  
            raise e
        finally:
            self.learn.path = temp
            self.learn.model_dir = 'models'

        if framework.lower() == "tf-onnx":
            batch_size = kwargs.get('batch_size', 16)

            with nostdout():
                self._save_as_tfonnx(saved_path, batch_size)
                zip_name = self._create_tfonnx_emd(saved_path.with_suffix('.onnx'), batch_size)
                os.remove(saved_path.with_suffix('.pth'))
        else:
            zip_name = self._create_emd(saved_path)

            if save_html:
                self._save_model_characteristics(saved_path.parent.absolute()/model_characteristics_folder)
                ArcGISModel._create_html(saved_path)
                
        if self._emd_template.get('InferenceFunction', False):
            with open(saved_path.parent / self._emd_template['InferenceFunction'], 'w') as f:
                f.write(self._code)
        if zip_files:
            _create_zip(str(zip_name), str(saved_path.parent))
        if arcgis.env.verbose:
            print('Created model files at {spp}'.format(spp=saved_path.parent))

        if publish:
            self._publish_dlpk((saved_path.parent/saved_path.stem).with_suffix('.dlpk'), gis=gis)

        return saved_path.parent

    def _save_model_characteristics(self, model_characteristics_dir):
        import matplotlib.pyplot as plt

        if not os.path.exists(os.path.join(model_characteristics_dir, model_characteristics_dir)):
            os.mkdir(os.path.join(model_characteristics_dir, model_characteristics_dir))

        if hasattr(self.learn, 'recorder'):
            self.learn.recorder.plot_losses()
            plt.savefig(os.path.join(model_characteristics_dir, 'loss_graph.png'))
            plt.close()
        self.show_results()
        plt.savefig(os.path.join(model_characteristics_dir, 'show_results.png'))
        plt.close()

        if hasattr(self, '_save_confusion_matrix'):
            self._save_confusion_matrix(model_characteristics_dir)

    def _publish_dlpk(self, dlpk_path, gis=None):
        gis_user = arcgis.env.active_gis if gis is None else gis
        if not gis_user:
            warn('No active gis user found!')
            return

        if not os.path.exists(dlpk_path):
            warn('DLPK file not found!')
            return

        emd_path = os.path.join(dlpk_path.parent, dlpk_path.stem + '.emd')

        if not os.path.exists(emd_path):
            warn('EMD File not found!')
            return

        emd_data = json.load(open(emd_path, 'r'))
        formatted_description = f"""
                <p><b> {emd_data.get('ModelName').replace('>', '').replace('<', '')} </b></p>
                <p><b>Backbone:</b> {emd_data.get('ModelParameters', {}).get('backbone')}</p>
                <p><b>Learning Rate:</b> {emd_data.get('LearningRate')}</p>
        """

        if emd_data.get('accuracy'):
            formatted_description = formatted_description + f"""
                <p><b>Analysis of the model</b></p>
                <p><b>Accuracy:</b> {emd_data.get('accuracy')}</p>
            """

        if emd_data.get('average_precision_score'):
            formatted_description = formatted_description + f"""
                <p><b>Analysis of the model</b></p>
                <p><b>Average Precision Score:</b> {emd_data.get('average_precision_score')}</p>
            """

        item = gis_user.content.add(
            {'type': 'Deep Learning Package', 'description': formatted_description, 'title': dlpk_path.stem},
            data=str(dlpk_path.absolute())
        )

        print(f"Published DLPK Item Id: {item.itemid}")

        model_characteristics_dir = os.path.join(dlpk_path.parent.absolute(), model_characteristics_folder)
        screenshots = [os.path.join(model_characteristics_dir, screenshot) for screenshot in os.listdir(model_characteristics_dir)]

        item.update(item_properties={'screenshots': screenshots})

    def _create_tfonnx_emd(self, saved_path, batch_size):
        "Raises error if framework specified is TF-ONNX but is not supported by the model"
        raise NotImplementedError('TF-ONNX framework is currently not supported by this model.')

    def _save_as_tfonnx(self, saved_path, batch_size):
        try:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                import onnx
                from onnx_tf.backend import prepare
        except:
            raise Exception('Tensorflow(version 1.13.1 or above), Onnx(version 1.5.0) and Onnx_tf(version 1.3.0) libraries are not installed. Install Tensorflow using "conda install tensorflow-gpu=1.13.1". Install onnx and onnx_tf using "pip install onnx onnx_tf".')

        batch_size = int(math.sqrt(int(batch_size)))**2
        dummy_input = torch.randn(batch_size, 3, self._data.chip_size, self._data.chip_size, device=self._device, requires_grad=True)
        torch.onnx.export(self.learn.model, dummy_input, saved_path.with_suffix('.onnx'))

    def save(self, name_or_path, framework='PyTorch', publish=False, gis=None, **kwargs):
        """
        Saves the model weights, creates an Esri Model Definition and Deep
        Learning Package zip for deployment to Image Server or ArcGIS Pro.   
        
        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        name_or_path            Required string. Name of the model to save. It
                                stores it at the pre-defined location. If path
                                is passed then it stores at the specified path
                                with model name as directory name and creates
                                all the intermediate directories.
        ---------------------   -------------------------------------------
        framework               Optional string. Defines the framework of the
                                model. (Only supported by ``SingleShotDetector``, currently.)
                                If framework used is ``TF-ONNX``, ``batch_size`` is required
                                to be passed as keyword arguments. 
                                
                                Choice list: ['PyTorch', 'TF-ONNX']
        ---------------------   -------------------------------------------
        publish                 Optional boolean. Publishes the DLPK as an item.
        ---------------------   -------------------------------------------
        gis                     Optional GIS Object. Used for publishing the item.
                                If not specified then active gis user is taken.
        =====================   ===========================================
        """        
        return self._save(name_or_path, framework=framework, publish=publish, gis=gis, **kwargs)
        
    def load(self, name_or_path):
        """
        Loads a saved model for inferencing or fine tuning from the specified
        path or model name.
        
        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        name_or_path            Required string. Name of the model to load from
                                the pre-defined location. If path is passed then
                                it loads from the specified path with model name
                                as directory name. Path to ".pth" file can also
                                be passed
        =====================   ===========================================
        """
        temp = self.learn.path
        if '\\' in name_or_path or '/' in name_or_path:
            path = Path(name_or_path)
            # to make fastai from both path and with name
            if path.is_file():
                name = path.stem
                self.learn.path = path.parent
            else:
                name = path.parts[-1]
                self.learn.path = path
            self.learn.model_dir = ''
        else:
            # fixing fastai bug
            self.learn.path = self.learn.path.parent
            self.learn.model_dir =  Path(self.learn.model_dir) /  name_or_path
            name = name_or_path

        try:
            self.learn.load(name, purge=False)
        except Exception as e:
            raise e
        finally:
            # undoing changes to self.learn.path
            self.learn.path = temp
            self.learn.model_dir = 'models'



from .._layer import ImageryLayer
from arcgis.gis import Item
import numbers
from arcgis.features.layer import FeatureLayer

def _raster_input(raster, raster2=None):
    layer=None
    if raster2 is not None:
        if isinstance(raster2, ImageryLayer) and isinstance(raster, ImageryLayer):
            layer = raster2
            raster_ra = _get_raster_ra(raster2)
            if raster._datastore_raster and raster2._datastore_raster:
                if raster2._fn is not None:
                    if raster._uri == raster2._uri:
                        raster2 = raster2._fn
                    else:
                        raster2 = _replace_raster_url(raster2._fn, raster2._uri)
                else:
                    if raster2._uri == raster._uri:
                        oids = raster2.filtered_rasters()
                        if oids is None:
                            raster2 = '$$'
                        elif len(oids) == 1:
                            raster2 = '$' + str(oids[0])
                        else:
                            raster2 = ['$' + str(x) for x in oids]
                    else:
                        raster2 = raster2._uri
            else:
                 if raster2._fn is not None:
                    if raster._url == raster2._url:
                        raster2 = raster2._fn
                    else:
                        if(raster2._datastore_raster is False):
                            if "url" in raster2._lyr_dict:
                                url = raster2._lyr_dict["url"]
                                if "serviceToken" in raster2._lyr_dict:
                                    url = url+"?token="+ raster2._lyr_dict["serviceToken"]
                                raster2 = _replace_raster_url(raster2._fn, url)
                            else:
                                raster2 = _replace_raster_url(raster2._fn, raster2._url)
                        else:
                            raster2 = _replace_raster_url(raster2._fn, raster2._uri)
                 else:
                    if raster2._url == raster._url:
                        oids = raster2.filtered_rasters()
                        if oids is None:
                            raster2 = '$$'
                        elif len(oids) == 1:
                            raster2 = '$' + str(oids[0])
                        else:
                            raster2 = ['$' + str(x) for x in oids]
                    else:
                        if(raster2._datastore_raster is False):
                            if "url" in raster2._lyr_dict:
                                url = raster2._lyr_dict["url"]
                                if "serviceToken" in raster2._lyr_dict:
                                    url = url+"?token="+ raster2._lyr_dict["serviceToken"]
                                raster2 = url
                            else:
                                raster2 = raster2._url
                        else:
                            raster2 = raster2._uri
        elif isinstance(raster2, ImageryLayer) and not isinstance(raster, ImageryLayer):
            layer = raster2
            raster_ra = _get_raster_ra(raster2)
            raster2 = _get_raster(raster2)

        elif isinstance(raster2, list):
            mix_and_match = False # mixing rasters from two image services
            # try:
            #     r0 = raster2[0]
            #     r1 = raster2[1]
            #     if r0._fn is None and r1._fn is None and r0._url != r1._url:
            #         mix_and_match = True
            # except:
            #     pass

            for r in raster2: # layer is first non numeric raster in list
                if not isinstance(r, numbers.Number):
                    layer = r
                    break

            for r in raster2:
                if r._datastore_raster:
                    if not isinstance(r, numbers.Number):
                        if r._uri != layer._uri:
                            mix_and_match = True
                else:
                    if not isinstance(r, numbers.Number):
                        if r._url != layer._url:
                            mix_and_match = True

            raster_ra = [_get_raster_ra(r) for r in raster2]
            if mix_and_match:
                raster2 = [_get_raster_url(r, layer) for r in raster2]
            else:
                raster2 = [_get_raster(r) for r in raster2]
        else: # secondinput maybe scalar for arithmetic functions, or a chained raster fn
            layer = None
            # raster = raster
            raster_ra = raster2
        return layer, raster2, raster_ra

    if isinstance(raster, ImageryLayer):
        layer = raster
        raster_ra = _get_raster_ra(raster)
        raster = _get_raster(raster)

    elif isinstance(raster, list):
        mix_and_match = False # mixing rasters from two image services
        # try:
        #     r0 = raster[0]
        #     r1 = raster[1]
        #     if r0._fn is None and r1._fn is None and r0._url != r1._url:
        #         mix_and_match = True
        # except:
        #     pass

        for r in raster:
            if isinstance(r, ImageryLayer):
                if r._datastore_raster:
                    layer = r
        if layer is None:
            for r in raster: # layer is first non numeric raster in list
                if not isinstance(r, numbers.Number):
                    layer = r
                    break

        for r in raster:
            if not isinstance(r, numbers.Number):
                if r._datastore_raster:
                    if r._uri != layer._uri:
                        mix_and_match = True
                else:
                    if r._url != layer._url:
                        mix_and_match = True

        raster_ra = [_get_raster_ra(r) for r in raster]
        if mix_and_match:
            raster = [_get_raster_url(r, layer) for r in raster]
        else:
            raster = [_get_raster(r) for r in raster]
    else: # maybe scalar for arithmetic functions, or a chained raster fn
        layer = None
        # raster = raster
        raster_ra = raster

    return layer, raster, raster_ra

def _get_raster(raster):
    if isinstance(raster, ImageryLayer):
        if raster._fn is not None:
            raster = raster._fn
        else:
            oids = raster.filtered_rasters()
            if oids is None:
                raster = '$$'
            elif len(oids) == 1:
                raster = '$' + str(oids[0])
            else:
                raster = ['$' + str(x) for x in oids]
    return raster


def _replace_raster_url(obj, url=None):
    # replace all "Raster" : '$$' with url
    if isinstance(obj, dict):
        value = {k: _replace_raster_url(v, url)
                 for k, v in obj.items()}
    elif isinstance(obj, list):
        value = [_replace_raster_url(elem, url)
                 for elem in obj]
    else:
        value = obj

    if value == '$$':
        return url
    elif isinstance(value, str) and len(value) > 0 and value[0] == '$':
        return url + '/' + value.replace('$', '')
    else:
        return value


def _get_raster_url(raster, layer):
    if isinstance(raster, ImageryLayer):
        if raster._fn is not None:
            if raster._datastore_raster and layer._datastore_raster:
                if raster._uri == layer._uri:
                    raster = raster._fn
                else:
                    raster = _replace_raster_url(raster._fn, raster._uri)

            else:
                if raster._url == layer._url:
                    raster = raster._fn
                else:
                    if "url" in raster._lyr_dict:
                        url = raster._lyr_dict["url"]
                        if "serviceToken" in raster._lyr_dict:
                            url = url+"?token="+ raster._lyr_dict["serviceToken"]
                        raster = _replace_raster_url(raster._fn, url)
                    else:
                        raster = _replace_raster_url(raster._fn, raster._url)

        else:
            if raster._datastore_raster and layer._datastore_raster:
                if raster._uri == layer._uri:
                    raster = '$$'
                else:
                    raster = raster._uri
            else:
                if raster._url == layer._url:
                    raster = '$$'
                else:
                    if "url" in raster._lyr_dict:
                        url = raster._lyr_dict["url"]
                        if "serviceToken" in raster._lyr_dict:
                            url = url+"?token="+ raster._lyr_dict["serviceToken"]
                        raster = url
                    else:
                        raster = raster._url

            # oids = raster.filtered_rasters()
            # if oids is None:
            #     raster = '$$'
            # elif len(oids) == 1:
            #     raster = '$' + str(oids[0])
            # else:
            #     raster = ['$' + str(x) for x in oids]
    return raster


def _get_raster_ra(raster):

    if isinstance(raster, ImageryLayer):
        if "url" in raster._lyr_dict:
            url = raster._lyr_dict["url"]
        if "serviceToken" in raster._lyr_dict:
            url = url+"?token="+ raster._lyr_dict["serviceToken"]
        if raster._fnra is not None:
            raster_ra = raster._fnra
        else:
            raster_ra = {}
            if raster._mosaic_rule is not None:
                if raster._datastore_raster:
                    raster_ra["uri"] = raster._uri
                else:
                    raster_ra["url"] = url
                raster_ra["mosaicRule"] = raster._mosaic_rule
            else:
                if raster._datastore_raster:
                    raster_ra = raster._uri
            
                else:
                    raster_ra = url

            #if raster._mosaic_rule is not None:
            #    raster_ra['mosaicRule'] = raster._mosaic_rule
    elif isinstance(raster, Item):
        raise RuntimeError('Item not supported as input. Use ImageryLayer - e.g. item.layers[0]')
        #raster_ra = {
        #    'itemId': raster.itemid
        #}
    else:
        raster_ra = raster

    return raster_ra


def _raster_input_rft(raster, raster2=None):
    if isinstance(raster, ImageryLayer) or isinstance(raster,FeatureLayer):
        raster_ra = _get_raster_ra_rft(raster)

    elif isinstance(raster, list):
        raster_ra = [_get_raster_ra_rft(r) for r in raster]

    else: # maybe scalar for arithmetic functions, or a chained raster fn
        raster_ra = raster

    return raster_ra


def _get_raster_ra_rft(raster):
    if isinstance(raster, ImageryLayer):
        if "url" in raster._lyr_dict:
            url = raster._lyr_dict["url"]
        if "serviceToken" in raster._lyr_dict:
            url = url+"?token="+ raster._lyr_dict["serviceToken"]
        if raster._fnra is not None:
            raster_ra = raster._fnra
        else:
            raster_ra = {}
            if raster._mosaic_rule is not None:
                if raster._datastore_raster:
                    raster_ra["uri"] = raster._uri
                else:
                    raster_ra["url"] = url
                raster_ra["mosaicRule"] = raster._mosaic_rule
            else:
                if raster._datastore_raster:
                    raster_ra = raster._uri
                else:
                    raster_ra = url


    elif isinstance(raster,FeatureLayer):
        raster_ra = raster._url
            #if raster._mosaic_rule is not None:
            #    raster_ra['mosaicRule'] = raster._mosaic_rule
    elif isinstance(raster, Item):
        raise RuntimeError('Item not supported as input. Use ImageryLayer - e.g. item.layers[0]')
    else:
        raster_ra = raster

    return raster_ra

def _input_rft(input_layer):
    if isinstance(input_layer, dict):
        input_param = input_layer

    elif isinstance(input_layer, str):
        if '/fileShares/' in input_layer or '/rasterStores/' in input_layer or '/cloudStores/' in input_layer:
            input_param = {"uri": input_layer}
        else:
            input_param = {"url": input_layer}

    elif isinstance(input_layer, list):
        input_param = []
        for il in input_layer:
            if isinstance(il,str):
                if '/fileShares/' in il or '/rasterStores/' in il or '/cloudStores/' in il:
                    input_param.append({"uri":il})
                else:
                    input_param.append({"url":il})
            elif isinstance(il,dict):
                input_param.append(il)
    else:
        input_param = input_layer
    return input_param




def _find_object_ref(rft_dict, record, instance):
    for k, v in rft_dict.items():
        if isinstance(v, dict):
            if "_object_id" in v:
                record[v["_object_id"]] = v
            if "isPublic" in v:
                if v["isPublic"] is True:
                    instance._is_public_flag=True
            _find_object_ref(v, record, instance)

        elif isinstance(v, list):
            for ele in v:
                if isinstance (ele,dict):
                    if "_object_id" in ele:
                        record[ele["_object_id"]] = ele
                    if "isPublic" in ele:
                        if ele["isPublic"] is True:
                            instance._is_public_flag=True
                    _find_object_ref(ele, record, instance)
    return _replace_object_id(rft_dict, record)

def _replace_object_id(rft_dict, record):

    if isinstance (rft_dict, dict):
        if "_object_ref_id" in rft_dict.keys():
                ref_value=record[rft_dict["_object_ref_id"]]
                rft_dict=ref_value
                return _replace_object_id(rft_dict, record)
        else:
            for k, v in rft_dict.items():
                if isinstance(v, dict):
                    rft_dict[k]=_replace_object_id(v, record)

                elif isinstance(v, list):
                    for n,ele in enumerate(v):
                        if isinstance (ele,dict):
                            v[n]=_replace_object_id(ele, record)
    return rft_dict


def _python_variable_name(var):
        var = ''.join(e for e in var if e.isalnum() or e=="_")
        return var

def _pixel_type_string_to_long(pixel_type):
    pixel_type = pixel_type.upper()
    if(pixel_type ==  'U1'):
        return 0
    elif(pixel_type ==  'U2'):
            return 1
    elif(pixel_type ==  'U4'):
            return 2
    elif(pixel_type ==  'U8'):
            return 3
    elif(pixel_type ==  'S8'):
            return 4
    elif(pixel_type ==  'U16'):
            return 5
    elif(pixel_type ==  'S16'):
            return 6
    elif(pixel_type ==  'U32'):
            return 7
    elif(pixel_type ==  'S32'):
            return 8
    elif(pixel_type ==  'F32'):
            return 9
    elif(pixel_type ==  'F64'):
            return 10
    elif(pixel_type ==  'C16'):
            return 13
    elif(pixel_type ==  'C32'):
            return 14
    elif(pixel_type ==  'C64'):
            return 11
    elif(pixel_type ==  'C128'):
            return 12
    elif(pixel_type ==  'D8'):
            return 15
    elif(pixel_type ==  'D10'):
            return 16
    elif(pixel_type ==  'D12'):
            return 17
    elif(pixel_type ==  'D16'):
            return 18
    elif(pixel_type ==  'D20'):
            return 19
    elif(pixel_type ==  'D24'):
            return 20
    elif(pixel_type ==  'D32'):
            return 21
    elif(pixel_type ==  'D36'):
            return 22
    else:
        return -1



from ._arcgis_model import ArcGISModel
import tempfile
from pathlib import Path
import json
from ._codetemplate import code
from ._arcgis_model import _EmptyData
import logging
from ._codetemplate import instance_detector_prf

try:
    import torch
    from fastai.vision.learner import cnn_learner
    from fastai.callbacks.hooks import model_sizes
    from fastai.vision.learner import create_body
    from fastai.vision.image import open_image
    from torchvision.models import resnet34
    from torchvision import models
    import numpy as np
    from .._data import prepare_data, _raise_fastai_import_error
    from fastai.callbacks import EarlyStoppingCallback
    from ._arcgis_model import SaveModelCallback, _set_multigpu_callback
    import torchvision
    from torchvision import models
    from torchvision.models.detection.faster_rcnn import FastRCNNPredictor
    from torchvision.models.detection.mask_rcnn import MaskRCNNPredictor
    from fastai.basic_train import Learner
    from ._maskrcnn_utils import is_no_color, mask_rcnn_loss, train_callback
    from fastai.torch_core import split_model_idx
    import matplotlib.pyplot as plt
    import matplotlib.patches as patches

    HAS_FASTAI = True
except Exception as e:
    #raise Exception(e)
    HAS_FASTAI = False


class MaskRCNN(ArcGISModel):
    """
    Creates a ``MaskRCNN`` Instance segmentation object

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    data                    Required fastai Databunch. Returned data object from
                            ``prepare_data`` function.
    ---------------------   -------------------------------------------
    backbone                Optional function. Backbone CNN model to be used for
                            creating the base of the `MaskRCNN`, which
                            is `resnet50` by default. 
                            Compatible backbones: 'resnet50'
    ---------------------   -------------------------------------------
    pretrained_path         Optional string. Path where pre-trained model is
                            saved.
    =====================   ===========================================

    :returns: ``MaskRCNN`` Object
    """
    def __init__(self, data, backbone=None, pretrained_path=None):

        super().__init__(data, backbone)
    
        self._backbone = models.resnet50

        #if not self._check_backbone_support(self._backbone):
        #    raise Exception (f"Enter only compatible backbones from {', '.join(self.supported_backbones)}")

        self._code = instance_detector_prf

        model = models.detection.maskrcnn_resnet50_fpn(pretrained=True, min_size = data.chip_size)
        in_features = model.roi_heads.box_predictor.cls_score.in_features
        model.roi_heads.box_predictor = FastRCNNPredictor(in_features, data.c)
        in_features_mask = model.roi_heads.mask_predictor.conv5_mask.in_channels
        hidden_layer = 256
        model.roi_heads.mask_predictor = MaskRCNNPredictor(in_features_mask,
                                                       hidden_layer,
                                                       data.c)

        self.learn = Learner(data, model, loss_func = mask_rcnn_loss)
        self.learn.callbacks.append(train_callback(self.learn))
        self.learn.model = self.learn.model.to(self._device)

        # fixes for zero division error when slice is passed
        self.learn.layer_groups = split_model_idx(self.learn.model, [28])
        self.learn.create_opt(lr=3e-3)

        if pretrained_path is not None:
            self.load(pretrained_path)           

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return '<%s>' % (type(self).__name__)

    @property
    def supported_backbones(self):
        return [models.detection.maskrcnn_resnet50_fpn.__name__]

    @classmethod
    def from_model(cls, emd_path, data=None):
        """
        Creates a ``MaskRCNN`` Instance segmentation object from an Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Esri Model Definition
                                file.
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from ``prepare_data`` function or None for
                                inferencing.

        =====================   ===========================================

        :returns: `MaskRCNN` Object
        """

        emd_path = Path(emd_path)
        with open(emd_path) as f:
            emd = json.load(f)
            
        model_file = Path(emd['ModelFile'])
        
        if not model_file.is_absolute():
            model_file = emd_path.parent / model_file
            
        model_params = emd['ModelParameters']

        try:
            class_mapping = {i['Value'] : i['Name'] for i in emd['Classes']}
            color_mapping = {i['Value'] : i['Color'] for i in emd['Classes']}
        except KeyError:
            class_mapping = {i['ClassValue'] : i['ClassName'] for i in emd['Classes']} 
            color_mapping = {i['ClassValue'] : i['Color'] for i in emd['Classes']}                

        
        if data is None:
            empty_data = _EmptyData(path=tempfile.TemporaryDirectory().name, loss_func=None, c=len(class_mapping) + 1, chip_size=emd['ImageHeight'])
            empty_data.class_mapping = class_mapping
            empty_data.color_mapping = color_mapping
            return cls(empty_data, **model_params, pretrained_path=str(model_file))
        else:
            return cls(data, **model_params, pretrained_path=str(model_file))        

    def _create_emd(self, path):
        import random
        super()._create_emd(path)
        self._emd_template["Framework"] = "arcgis.learn.models._inferencing"
        self._emd_template["ModelConfiguration"] = "_maskrcnn_inferencing"
        self._emd_template["InferenceFunction"] = "ArcGISInstanceDetector.py"

        self._emd_template["ExtractBands"] = [0, 1, 2]
        self._emd_template['Classes'] = []
        class_data = {}
        for i, class_name in enumerate(self._data.classes[1:]):  # 0th index is background
            inverse_class_mapping = {v: k for k, v in self._data.class_mapping.items()}
            class_data["Value"] = inverse_class_mapping[class_name]
            class_data["Name"] = class_name
            color = [random.choice(range(256)) for i in range(3)] if is_no_color(self._data.color_mapping) else \
            self._data.color_mapping[inverse_class_mapping[class_name]]
            class_data["Color"] = color
            self._emd_template['Classes'].append(class_data.copy())
            
        json.dump(self._emd_template, open(path.with_suffix('.emd'), 'w'), indent=4)
        return path.stem

    @property
    def _model_metrics(self):
        return {}

    def _predict_results(self, xb):

        self.learn.model.eval()
        predictions = self.learn.model(xb.cuda())
        predictionsf =[]
        for i in range(len(predictions)):
            predictionsf.append({})
            predictionsf[i]['masks'] = predictions[i]['masks'].detach().cpu().numpy()
            predictionsf[i]['boxes'] = predictions[i]['boxes'].detach().cpu().numpy()
            predictionsf[i]['labels'] = predictions[i]['labels'].detach().cpu().numpy()
            predictionsf[i]['scores'] = predictions[i]['scores'].detach().cpu().numpy()
            del predictions[i]['masks']
            del predictions[i]['boxes']
            del predictions[i]['labels']
            del predictions[i]['scores']
        del xb
        torch.cuda.empty_cache()      

        return predictionsf

    def _predict_postprocess(self, predictions, threshold=0.5, box_threshold = 0.5):

        pred_mask = []
        pred_box = []

        for i in range(len(predictions)):
            out = predictions[i]['masks'].squeeze()
            pred_box.append([])

            if out.shape[0] != 0:  # handle for prediction with n masks
                if len(out.shape) == 2: # for out dimension hxw (in case of only one predicted mask)
                    out = out[None]
                ymask = np.where(out[0]> threshold, 1, 0)
                #if torch.max(out[0]) > threshold:
                if predictions[i]['scores'][0] > box_threshold:
                    pred_box[i].append(predictions[i]['boxes'][0])
                for j in range(1,out.shape[0]):
                    ym1 = np.where(out[j]> threshold, j+1, 0)
                    ymask += ym1
                    #if torch.max(out[j]) > threshold:
                    if predictions[i]['scores'][j] > box_threshold:
                        pred_box[i].append(predictions[i]['boxes'][j])
            else:
                ymask = np.zeros((self._data.chip_size, self._data.chip_size)) # handle for not predicted masks
            pred_mask.append(ymask)
        return pred_mask, pred_box

    def show_results(self, mode='mask', mask_threshold=0.5, box_threshold=0.7, nrows=None, imsize=5, index=0, alpha=0.5, cmap='tab20'):
        """
        Displays the results of a trained model on a part of the validation set.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        mode                    Required arguments within ['bbox', 'mask', 'bbox_mask'].
                                    * ``bbox`` - For visualizing only boundig boxes.
                                    * ``mask`` - For visualizing only mask
                                    * ``bbox_mask`` - For visualizing both mask and bounding boxes.
        ---------------------   -------------------------------------------
        mask_threshold          Optional float. The probabilty above which
                                a pixel will be considered mask.
        ---------------------   -------------------------------------------
        box_threshold           Optional float. The pobabilty above which
                                a detection will be considered valid.
        ---------------------   -------------------------------------------
        nrows                   Optional int. Number of rows of results
                                to be displayed.
        =====================   ===========================================
        """ 

        if mode not in ['bbox', 'mask', 'bbox_mask']:
            raise Exception("mode can be only ['bbox', 'mask', 'bbox_mask']")

        # Get Number of items
        if nrows is None:
            nrows = self._data.batch_size
        ncols=2
    
        # Get Batch
        xb,yb = self._data.one_batch('DatasetType.Valid')
        
        predictions = self._predict_results(xb)

        pred_mask, pred_box = self._predict_postprocess(predictions, mask_threshold, box_threshold)

        fig, ax = plt.subplots(nrows=nrows, ncols=ncols, figsize=(ncols*imsize, nrows*imsize))
        fig.suptitle('Ground Truth / Predictions', fontsize=20)

        for i in range(nrows):
            ax[i][0].imshow(xb[i].numpy().transpose(1,2,0))
            ax[i][0].axis('off')
            if mode in ['mask', 'bbox_mask']:
                yb_mask = yb[i][0].numpy()
                for j in range(1, yb[i].shape[0]):
                    max_unique = np.max(np.unique(yb_mask))
                    yb_j = np.where(yb[i][j]>0, yb[i][j] + max_unique, yb[i][j])
                    yb_mask += yb_j
                ax[i][0].imshow(yb_mask, cmap = cmap, alpha = alpha)
            ax[i][0].axis('off')
            ax[i][1].imshow(xb[i].numpy().transpose(1,2,0))
            ax[i][1].axis('off')
            if mode in ['mask', 'bbox_mask']:
                ax[i][1].imshow(pred_mask[i], cmap=cmap, alpha = alpha)
            if mode in ['bbox','bbox']:
                if pred_box[i] != []:
                    for num_boxes in pred_box[i]:
                        rect = patches.Rectangle((num_boxes[0], num_boxes[1]), num_boxes[2]-num_boxes[0], num_boxes[3]-num_boxes[1], linewidth=1, edgecolor='r', facecolor='none')
                        ax[i][1].add_patch(rect)
            ax[i][1].axis('off')
        plt.subplots_adjust(top=0.95)
        torch.cuda.empty_cache()
